/*
/
/ spatialite.c - main container for SpatiaLite - spatial extension for SQLite
/
/ Author: Sandro Furieri a-furieri@lqt.it
/
/ Copyright (C) 2007  Alessandro Furieri
/
/ This library is free software; you can redistribute it and/or
/ modify it under the terms of the GNU Lesser General Public
/ License as published by the Free Software Foundation; either
/ version 2.1 of the License, or (at your option) any later version.
/
/ This library is distributed in the hope that it will be useful,
/ but WITHOUT ANY WARRANTY; without even the implied warranty of
/ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
/ Lesser General Public License for more details.
/
/ You should have received a copy of the GNU Lesser General Public
/ License along with this library; if not, write to the Free Software
/ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA/
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <float.h>
#include <math.h>
#include <proj_api.h>
#include "sqlite3ext.h"
#include "spatialite.h"

sqlite3 *this_db;

static void sple_get_proj_params (int srid, char *proj_params);
static int sple_is_longlat (char *proj_params);
static double sple_rads_to_degs (double rads);
static double sple_degs_to_rads (double degrees);

SQLITE_EXTENSION_INIT1
    SPATIALITE_DECLARE int
    sqlite3_extension_init
    (sqlite3 * db, char **pzErrMsg, const sqlite3_api_routines * pApi)
{
/* SQLite invokes this routine once when it loads the extension. */
    SQLITE_EXTENSION_INIT2 (pApi) this_db = db;
    sqlite3_create_function (db, "LoadShapefile", -1, SQLITE_ANY, 0,
			     LoadShapefile, 0, 0);
    sqlite3_create_function (db, "ImportShapefile", -1, SQLITE_ANY, 0,
			     LoadShapefile, 0, 0);
    sqlite3_create_function (db, "DumpShapefile", -1, SQLITE_ANY, 0,
			     DumpShapefile, 0, 0);
    sqlite3_create_function (db, "ExportShapefile", -1, SQLITE_ANY, 0,
			     DumpShapefile, 0, 0);
    sqlite3_create_function (db, "AsText", 1, SQLITE_ANY, 0, AsText, 0, 0);
    sqlite3_create_function (db, "AsBinary", 1, SQLITE_ANY, 0, AsBinary, 0, 0);
    sqlite3_create_function (db, "GeomFromText", 1, SQLITE_ANY, 0,
			     GeomFromText1, 0, 0);
    sqlite3_create_function (db, "GeomFromText", 2, SQLITE_ANY, 0,
			     GeomFromText2, 0, 0);
    sqlite3_create_function (db, "GeometryFromText", 1, SQLITE_ANY, 0,
			     GeomFromText1, 0, 0);
    sqlite3_create_function (db, "GeometryFromText", 2, SQLITE_ANY, 0,
			     GeomFromText2, 0, 0);
    sqlite3_create_function (db, "GeomCollFromText", 1, SQLITE_ANY, 0,
			     GeomCollFromText1, 0, 0);
    sqlite3_create_function (db, "GeomCollFromText", 2, SQLITE_ANY, 0,
			     GeomCollFromText2, 0, 0);
    sqlite3_create_function (db, "GeometryCollectionFromText", 1, SQLITE_ANY, 0,
			     GeomCollFromText1, 0, 0);
    sqlite3_create_function (db, "GeometryCollectionFromText", 2, SQLITE_ANY, 0,
			     GeomCollFromText2, 0, 0);
    sqlite3_create_function (db, "LineFromText", 1, SQLITE_ANY, 0,
			     LineFromText1, 0, 0);
    sqlite3_create_function (db, "LineFromText", 2, SQLITE_ANY, 0,
			     LineFromText2, 0, 0);
    sqlite3_create_function (db, "LineStringFromText", 1, SQLITE_ANY, 0,
			     LineFromText1, 0, 0);
    sqlite3_create_function (db, "LineStringFromText", 2, SQLITE_ANY, 0,
			     LineFromText2, 0, 0);
    sqlite3_create_function (db, "MLineFromText", 1, SQLITE_ANY, 0,
			     MLineFromText1, 0, 0);
    sqlite3_create_function (db, "MLineFromText", 2, SQLITE_ANY, 0,
			     MLineFromText2, 0, 0);
    sqlite3_create_function (db, "MultiLineStringFromText", 1, SQLITE_ANY, 0,
			     MLineFromText1, 0, 0);
    sqlite3_create_function (db, "MultiLineStringFromText", 2, SQLITE_ANY, 0,
			     MLineFromText2, 0, 0);
    sqlite3_create_function (db, "MPointFromText", 1, SQLITE_ANY, 0,
			     MPointFromText1, 0, 0);
    sqlite3_create_function (db, "MPointFromText", 2, SQLITE_ANY, 0,
			     MPointFromText2, 0, 0);
    sqlite3_create_function (db, "MultiPointFromText", 1, SQLITE_ANY, 0,
			     MPointFromText1, 0, 0);
    sqlite3_create_function (db, "MultiPointFromText", 2, SQLITE_ANY, 0,
			     MPointFromText2, 0, 0);
    sqlite3_create_function (db, "MPolyFromText", 1, SQLITE_ANY, 0,
			     MPolyFromText1, 0, 0);
    sqlite3_create_function (db, "MPolyFromText", 2, SQLITE_ANY, 0,
			     MPolyFromText2, 0, 0);
    sqlite3_create_function (db, "MultiPolygonFromText", 1, SQLITE_ANY, 0,
			     MPolyFromText1, 0, 0);
    sqlite3_create_function (db, "MultiPolygonFromText", 2, SQLITE_ANY, 0,
			     MPolyFromText2, 0, 0);
    sqlite3_create_function (db, "PointFromText", 1, SQLITE_ANY, 0,
			     PointFromText1, 0, 0);
    sqlite3_create_function (db, "PointFromText", 2, SQLITE_ANY, 0,
			     PointFromText2, 0, 0);
    sqlite3_create_function (db, "PolyFromText", 1, SQLITE_ANY, 0,
			     PolyFromText1, 0, 0);
    sqlite3_create_function (db, "PolyFromText", 2, SQLITE_ANY, 0,
			     PolyFromText2, 0, 0);
    sqlite3_create_function (db, "PolygonFromText", 1, SQLITE_ANY, 0,
			     PolyFromText1, 0, 0);
    sqlite3_create_function (db, "PolygomFromText", 2, SQLITE_ANY, 0,
			     PolyFromText2, 0, 0);
    sqlite3_create_function (db, "GeomFromWKB", 1, SQLITE_ANY, 0,
			     GeomFromWKB1, 0, 0);
    sqlite3_create_function (db, "GeomFromWKB", 2, SQLITE_ANY, 0,
			     GeomFromWKB2, 0, 0);
    sqlite3_create_function (db, "GeometryFromWKB", 1, SQLITE_ANY, 0,
			     GeomFromWKB1, 0, 0);
    sqlite3_create_function (db, "GeometryFromWKB", 2, SQLITE_ANY, 0,
			     GeomFromWKB2, 0, 0);
    sqlite3_create_function (db, "GeomCollFromWKB", 1, SQLITE_ANY, 0,
			     GeomCollFromWKB1, 0, 0);
    sqlite3_create_function (db, "GeomCollFromWKB", 2, SQLITE_ANY, 0,
			     GeomCollFromWKB2, 0, 0);
    sqlite3_create_function (db, "GeometryCollectionFromWKB", 1, SQLITE_ANY, 0,
			     GeomCollFromWKB1, 0, 0);
    sqlite3_create_function (db, "GeometryCollectionFromWKB", 2, SQLITE_ANY, 0,
			     GeomCollFromWKB2, 0, 0);
    sqlite3_create_function (db, "LineFromWKB", 1, SQLITE_ANY, 0,
			     LineFromWKB1, 0, 0);
    sqlite3_create_function (db, "LineFromWKB", 2, SQLITE_ANY, 0,
			     LineFromWKB2, 0, 0);
    sqlite3_create_function (db, "LineStringFromWKB", 1, SQLITE_ANY, 0,
			     LineFromWKB1, 0, 0);
    sqlite3_create_function (db, "LineStringFromWKB", 2, SQLITE_ANY, 0,
			     LineFromWKB2, 0, 0);
    sqlite3_create_function (db, "MLineFromWKB", 1, SQLITE_ANY, 0,
			     MLineFromWKB1, 0, 0);
    sqlite3_create_function (db, "MLineFromWKB", 2, SQLITE_ANY, 0,
			     MLineFromWKB2, 0, 0);
    sqlite3_create_function (db, "MultiLineStringFromWKB", 1, SQLITE_ANY, 0,
			     MLineFromWKB1, 0, 0);
    sqlite3_create_function (db, "MultiLineStringFromWKB", 2, SQLITE_ANY, 0,
			     MLineFromWKB2, 0, 0);
    sqlite3_create_function (db, "MPointFromWKB", 1, SQLITE_ANY, 0,
			     MPointFromWKB1, 0, 0);
    sqlite3_create_function (db, "MPointFromWKB", 2, SQLITE_ANY, 0,
			     MPointFromWKB2, 0, 0);
    sqlite3_create_function (db, "MultiPointFromWKB", 1, SQLITE_ANY, 0,
			     MPointFromWKB1, 0, 0);
    sqlite3_create_function (db, "MultiPointFromWKB", 2, SQLITE_ANY, 0,
			     MPointFromWKB2, 0, 0);
    sqlite3_create_function (db, "MPolyFromWKB", 1, SQLITE_ANY, 0,
			     MPolyFromWKB1, 0, 0);
    sqlite3_create_function (db, "MPolyFromWKB", 2, SQLITE_ANY, 0,
			     MPolyFromWKB2, 0, 0);
    sqlite3_create_function (db, "MultiPolygonFromWKB", 1, SQLITE_ANY, 0,
			     MPolyFromWKB1, 0, 0);
    sqlite3_create_function (db, "MultiPolygomFromWKB", 2, SQLITE_ANY, 0,
			     MPolyFromWKB2, 0, 0);
    sqlite3_create_function (db, "PointFromWKB", 1, SQLITE_ANY, 0,
			     PointFromWKB1, 0, 0);
    sqlite3_create_function (db, "PointFromWKB", 2, SQLITE_ANY, 0,
			     PointFromWKB2, 0, 0);
    sqlite3_create_function (db, "PolyFromWKB", 1, SQLITE_ANY, 0,
			     PolyFromWKB1, 0, 0);
    sqlite3_create_function (db, "PolyFromWKB", 2, SQLITE_ANY, 0,
			     PolyFromWKB2, 0, 0);
    sqlite3_create_function (db, "PolygonFromWKB", 1, SQLITE_ANY, 0,
			     PolyFromWKB1, 0, 0);
    sqlite3_create_function (db, "PolygonFromWKB", 2, SQLITE_ANY, 0,
			     PolyFromWKB2, 0, 0);
    sqlite3_create_function (db, "Dimension", 1, SQLITE_ANY, 0, Dimension, 0,
			     0);
    sqlite3_create_function (db, "GeometryType", 1, SQLITE_ANY, 0, GeometryType,
			     0, 0);
    sqlite3_create_function (db, "SRID", 1, SQLITE_ANY, 0, Srid, 0, 0);
    sqlite3_create_function (db, "IsEmpty", 1, SQLITE_ANY, 0, IsEmpty, 0, 0);
    sqlite3_create_function (db, "Envelope", 1, SQLITE_ANY, 0, Envelope, 0, 0);
    sqlite3_create_function (db, "X", 1, SQLITE_ANY, 0, X, 0, 0);
    sqlite3_create_function (db, "Y", 1, SQLITE_ANY, 0, Y, 0, 0);
    sqlite3_create_function (db, "IsClosed", 1, SQLITE_ANY, 0, IsClosed, 0, 0);
    sqlite3_create_function (db, "IsSimple", 1, SQLITE_ANY, 0, IsSimple, 0, 0);
    sqlite3_create_function (db, "IsRing", 1, SQLITE_ANY, 0, IsRing, 0, 0);
    sqlite3_create_function
	(db, "NumPoints", 1, SQLITE_ANY, 0, NumPoints, 0, 0);
    sqlite3_create_function
	(db, "StartPoint", 1, SQLITE_ANY, 0, StartPoint, 0, 0);
    sqlite3_create_function (db, "EndPoint", 1, SQLITE_ANY, 0, EndPoint, 0, 0);
    sqlite3_create_function (db, "PointN", 2, SQLITE_ANY, 0, PointN, 0, 0);
    sqlite3_create_function (db, "ExteriorRing", 1, SQLITE_ANY, 0, ExteriorRing,
			     0, 0);
    sqlite3_create_function (db, "NumInteriorRings", 1, SQLITE_ANY, 0,
			     NumInteriorRings, 0, 0);
    sqlite3_create_function (db, "NumInteriorRing", 1, SQLITE_ANY, 0,
			     NumInteriorRings, 0, 0);
    sqlite3_create_function (db, "InteriorRingN", 2, SQLITE_ANY, 0,
			     InteriorRingN, 0, 0);
    sqlite3_create_function (db, "NumGeometries", 1, SQLITE_ANY, 0,
			     NumGeometries, 0, 0);
    sqlite3_create_function (db, "GeometryN", 2, SQLITE_ANY, 0, GeometryN, 0,
			     0);
    sqlite3_create_function (db, "GLength", 1, SQLITE_ANY, 0, Length, 0, 0);
    sqlite3_create_function (db, "Area", 1, SQLITE_ANY, 0, Area, 0, 0);
    sqlite3_create_function (db, "Centroid", 1, SQLITE_ANY, 0, Centroid, 0, 0);
    sqlite3_create_function (db, "MBRContains", 2, SQLITE_ANY, 0, MBRContains,
			     0, 0);
    sqlite3_create_function (db, "MBRDisjoint", 2, SQLITE_ANY, 0, MBREqual, 0,
			     0);
    sqlite3_create_function (db, "MBREqual", 2, SQLITE_ANY, 0, MBREqual, 0, 0);
    sqlite3_create_function (db, "MBRIntersects", 2, SQLITE_ANY, 0,
			     MBRIntersects, 0, 0);
    sqlite3_create_function (db, "MBROverlaps", 2, SQLITE_ANY, 0, MBROverlaps,
			     0, 0);
    sqlite3_create_function (db, "MBRTouches", 2, SQLITE_ANY, 0, MBRTouches, 0,
			     0);
    sqlite3_create_function (db, "MBRWithin", 2, SQLITE_ANY, 0, MBRWithin, 0,
			     0);
    sqlite3_create_function (db, "ShiftCoords", 3, SQLITE_ANY, 0, ShiftCoords,
			     0, 0);
    sqlite3_create_function (db, "ShiftCoordinates", 3, SQLITE_ANY, 0,
			     ShiftCoords, 0, 0);
    sqlite3_create_function (db, "ScaleCoords", 2, SQLITE_ANY, 0, ScaleCoords,
			     0, 0);
    sqlite3_create_function (db, "ScaleCoords", 3, SQLITE_ANY, 0, ScaleCoords,
			     0, 0);
    sqlite3_create_function (db, "RotateCoords", 2, SQLITE_ANY, 0, RotateCoords,
			     0, 0);
    sqlite3_create_function (db, "ReflectCoords", 3, SQLITE_ANY, 0,
			     ReflectCoords, 0, 0);
    sqlite3_create_function (db, "ScaleCoordinates", 2, SQLITE_ANY, 0,
			     ScaleCoords, 0, 0);
    sqlite3_create_function (db, "ScaleCoordinates", 3, SQLITE_ANY, 0,
			     ScaleCoords, 0, 0);
    sqlite3_create_function (db, "RotateCoordinates", 2, SQLITE_ANY, 0,
			     RotateCoords, 0, 0);
    sqlite3_create_function (db, "ReflectCoordinates", 3, SQLITE_ANY, 0,
			     ReflectCoords, 0, 0);
    sqlite3_create_function (db, "Transform", 2, SQLITE_ANY, 0,
			     Transform, 0, 0);
    sqlite3_create_function (db, "SetSrid", 2, SQLITE_ANY, 0, SetSrid, 0, 0);
    return 0;
}

SPATIALITE_DECLARE void
LoadShapefile (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* pseudo SQL function:
/ LoadShapefile(path, table)
/
/ returns a strings explaing what really happened
*/
    int err_path = 0;
    int err_table = 0;
    int err_srid = 0;
    int err_column = 0;
    const unsigned char *shp_path = NULL;
    const unsigned char *table_name = NULL;
    const unsigned char *column_name = NULL;
    int srid = -1;
    char *error = NULL;
    int ret;
    char dummy[512];
    int len = 4096;
    char *p_result = malloc (len);
    int entities = 0;
    int shape = 0;
    char *class;
    clock_t t0;
    clock_t t1;
    clock_t diff_t;
    clock_t secs;
    clock_t millis;
    if (argc < 2 || argc > 4)
      {
	  strcpy (p_result, "\nERROR: wrong arguments !!!\n\n");
	  strcat (p_result, "the expected syntax is:\n");
	  strcat (p_result,
		  "LoadShapefile('path_to_shapefile', 'sqlite_table_name' [, srid] [, 'geocolumn_name']);\n\n");
	  sqlite3_result_text (context, p_result, len, free);
	  return;
      }
    if (sqlite3_value_type (argv[0]) != SQLITE_TEXT)
	err_path = 1;
    else
	shp_path = sqlite3_value_text (argv[0]);
    if (sqlite3_value_type (argv[1]) != SQLITE_TEXT)
	err_table = 1;
    else
	table_name = sqlite3_value_text (argv[1]);
    if (argc == 3)
      {
	  if (sqlite3_value_type (argv[2]) != SQLITE_INTEGER)
	      err_srid = 1;
	  else
	      srid = sqlite3_value_int (argv[2]);
      }
    else
	srid = -1;
    if (argc == 4)
      {
	  if (sqlite3_value_type (argv[3]) != SQLITE_TEXT)
	      err_column = 1;
	  else
	      column_name = sqlite3_value_text (argv[2]);
      }
    else
	column_name = (unsigned char *) ("geom");
    if (err_path || err_table || err_srid || err_column)
      {
	  strcpy (p_result, "\nERROR: wrong arguments !!!\n\n");
	  strcat (p_result, "the expected syntax is:\n");
	  strcat (p_result,
		  "LoadShapefile('path_to_shapefile', 'sqlite_table_name' [, srid] [, 'geocolumn_name']);\n\n");
      }
    else
      {
	  sprintf (p_result, "\n========\nLoading shapefile from '%s'\n",
		   shp_path);
	  sprintf (dummy, "into SQLite table '%s'\n", table_name);
	  strcat (p_result, dummy);
	  t0 = clock ();
	  ret =
	      sple_load_shapefile (shp_path, table_name, column_name, srid,
				   &entities, &shape, &error);
	  t1 = clock ();
	  if (ret < 0)
	    {
		sprintf (dummy, "========\nERROR: %s\n\n", error);
		free (error);
		strcat (p_result, dummy);
	    }
	  else
	    {
		sprintf (dummy, "========\ntable '%s' created\n", table_name);
		strcat (p_result, dummy);
		if (shape == 1 || shape == 11 || shape == 21)
		    class = "POINT";
		else if (shape == 3 || shape == 13 || shape == 23)
		    class = "LINESTRING/MULTILINESTRING";
		else if (shape == 5 || shape == 15 || shape == 25)
		    class = "POLYGON/MULTIPOLYGON";
		else
		    class = "UNKNOWN";
		sprintf (dummy, "acquired %d %s entit%s from shapefile\n",
			 entities, class, (entities == 1) ? "y" : "ies");
		strcat (p_result, dummy);
		diff_t = t1 - t0;
		secs = (diff_t / CLOCKS_PER_SEC);
		millis = diff_t - (secs * CLOCKS_PER_SEC);
		sprintf (dummy, "inserted %d row%c\t\t[%ld.%03ld sec]\n", ret,
			 (ret == 1) ? ' ' : 's', secs, millis);
		strcat (p_result, dummy);
	    }
      }
    len = strlen (p_result);
    sqlite3_result_text (context, p_result, len, free);
}

SPATIALITE_DECLARE void
DumpShapefile (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* pseudo SQL function:
/ DumpShapefile(table, path)
/
/ returns a strings explaing what really happened
*/
    int err_path = 0;
    int err_table = 0;
    int err_class = 0;
    int err_column = 0;
    const unsigned char *shp_path = NULL;
    const unsigned char *table_name = NULL;
    const unsigned char *shp_class = NULL;
    const unsigned char *column_name = NULL;
    char *error = NULL;
    int shape = -1;
    char dummy[512];
    int len = 4096;
    char *class;
    int ret;
    int rows = 0;
    clock_t t0;
    clock_t t1;
    clock_t diff_t;
    clock_t secs;
    clock_t millis;
    char *p_result = malloc (len);
    if (argc < 3 || argc > 4)
      {
	  strcpy (p_result, "\nERROR: wrong arguments !!!\n\n");
	  strcat (p_result, "the expected syntax is:\n");
	  strcat (p_result,
		  "DumpShapefile('table', {'P'|'L'|'A'}', 'path_to_shapefile', ['geocolumn']);\n\n");
	  sqlite3_result_text (context, p_result, len, free);
	  return;
      }
    if (sqlite3_value_type (argv[0]) != SQLITE_TEXT)
	err_table = 1;
    else
	table_name = sqlite3_value_text (argv[0]);
    if (sqlite3_value_type (argv[1]) != SQLITE_TEXT)
	err_class = 1;
    else
      {
	  shp_class = sqlite3_value_text (argv[1]);
	  len = sqlite3_value_bytes (argv[1]);
	  if (len == 1)
	    {
		if (*shp_class == 'P' || *shp_class == 'p' || *shp_class == '1')
		    shape = SPLE_POINT;
		if (*shp_class == 'M' || *shp_class == 'm' || *shp_class == '8')
		    shape = SPLE_MULTIPOINT;
		if (*shp_class == 'L' || *shp_class == 'l' || *shp_class == '2')
		    shape = SPLE_LINESTRING;
		if (*shp_class == 'A' || *shp_class == 'a' || *shp_class == '3')
		    shape = SPLE_POLYGON;
	    }
	  if (len == 4)
	    {
		if (strcasecmp ((char *) shp_class, "LINE") == 0)
		    shape = SPLE_LINESTRING;
		if (strcasecmp ((char *) shp_class, "AREA") == 0)
		    shape = SPLE_POLYGON;
	    }
	  if (len == 5)
	    {
		if (strcasecmp ((char *) shp_class, "POINT") == 0)
		    shape = SPLE_POINT;
		if (strcasecmp ((char *) shp_class, "LINES") == 0)
		    shape = SPLE_LINESTRING;
		if (strcasecmp ((char *) shp_class, "AREAA") == 0)
		    shape = SPLE_POLYGON;
	    }
	  if (len == 6)
	    {
		if (strcasecmp ((char *) shp_class, "POINTS") == 0)
		    shape = SPLE_POINT;
	    }
	  if (len == 7)
	    {
		if (strcasecmp ((char *) shp_class, "POLYGON") == 0)
		    shape = SPLE_POLYGON;
	    }
	  if (len == 8)
	    {
		if (strcasecmp ((char *) shp_class, "POLYGONS") == 0)
		    shape = SPLE_POLYGON;
	    }
	  if (len == 10)
	    {
		if (strcasecmp ((char *) shp_class, "LINESTRING") == 0)
		    shape = SPLE_LINESTRING;
		if (strcasecmp ((char *) shp_class, "MULTIPOINT") == 0)
		    shape = SPLE_MULTIPOINT;
	    }
	  if (len == 11)
	    {
		if (strcasecmp ((char *) shp_class, "LINESTRINGS") == 0)
		    shape = SPLE_LINESTRING;
		if (strcasecmp ((char *) shp_class, "MULTIPOINTS") == 0)
		    shape = SPLE_MULTIPOINT;
	    }
	  if (shape < 0)
	      err_class = 1;
      }
    if (sqlite3_value_type (argv[2]) != SQLITE_TEXT)
	err_path = 1;
    else
	shp_path = sqlite3_value_text (argv[2]);
    if (argc == 4)
      {
	  if (sqlite3_value_type (argv[3]) != SQLITE_TEXT)
	      err_column = 1;
	  else
	      column_name = sqlite3_value_text (argv[3]);
      }
    else
	column_name = (unsigned char *) ("geom");
    if (err_path || err_class || err_table || err_column)
      {
	  strcpy (p_result, "\nERROR: wrong arguments !!!\n\n");
	  strcat (p_result, "the expected syntax is:\n");
	  strcat (p_result,
		  "DumpPathname('table', {'1'|'2'|'3'|'8'}, 'path_to_shapefile' [, 'geocolumn']);\n\n");
      }
    else
      {
	  sprintf (p_result, "\n========\nDumping SQLite table '%s'\n",
		   table_name);
	  sprintf (dummy, "into shapefile at '%s'\n", shp_path);
	  strcat (p_result, dummy);
	  t0 = clock ();
	  ret =
	      sple_dump_shapefile (shp_path, table_name, column_name, &rows,
				   shape, &error);
	  t1 = clock ();
	  if (ret < 0)
	    {
		sprintf (dummy, "========\nERROR: %s\n\n", error);
		free (error);
		strcat (p_result, dummy);
	    }
	  else
	    {
		if (shape == SPLE_POINT)
		    class = "POINT";
		else if (shape == SPLE_LINESTRING)
		    class = "LINESTRING/MULTILINESTRING";
		else if (shape == SPLE_POLYGON)
		    class = "POLYGON/MULTIPOLYGON";
		else
		    class = "UNKNOWN";
		sprintf (dummy, "%d %s ro%s fetched from table\n", rows, class,
			 (rows == 1) ? "w" : "ws");
		strcat (p_result, dummy);
		diff_t = t1 - t0;
		secs = (diff_t / CLOCKS_PER_SEC);
		millis = diff_t - (secs * CLOCKS_PER_SEC);
		sprintf (dummy,
			 "%d entit%s exported to shapefile\t\t[%ld.%03ld sec]\n",
			 ret, (ret == 1) ? "y" : "ies", secs, millis);
		strcat (p_result, dummy);
	    }
      }
    len = strlen (p_result);
    sqlite3_result_text (context, p_result, len, free);
}

SPATIALITE_DECLARE void
AsText (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ AsText(BLOB encoded geometry)
/
/ returns the corresponding WKT encoded value
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int len;
    char *p_result = NULL;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  sple_out_wkt (geo, &p_result);
	  if (!p_result)
	      sqlite3_result_null (context);
	  else
	    {
		len = strlen (p_result);
		sqlite3_result_text (context, p_result, len, free);
	    }
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
AsBinary (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ AsBinary(BLOB encoded geometry)
/
/ returns the corresponding WKB encoded value
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int len;
    unsigned char *p_result = NULL;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  sple_blob_to_wkb (geo, &p_result, &len);
	  if (!p_result)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_blob (context, p_result, len, free);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
Dimension (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ Dimension(BLOB encoded geometry)
/
/ returns:
/ 0 if geometry is a POINT or MULTIPOINT
/ 1 if geometry is a LINESTRING or MULTILINESTRING
/ 2 if geometry is a POLYGON or MULTIPOLYGON
/ 0, 1, 2, for GEOMETRYCOLLECTIONS according to geoemetries contained inside
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int dim;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  dim = sple_dimension (geo);
	  sqlite3_result_int (context, dim);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
GeometryType (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ GeometryType(BLOB encoded geometry)
/
/ returns the class for current geometry:
/ 'POINT' or 'MULTIPOINT'
/ 'LINESTRING' or 'MULTILINESTRING'
/ 'POLYGON or MULTIPOLYGON
/ 'GEOMETRYCOLLECTIONS' 
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int len;
    char *p_result = NULL;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  sple_geometry_type (geo, &p_result);
	  if (!p_result)
	      sqlite3_result_null (context);
	  else
	    {
		len = strlen (p_result);
		sqlite3_result_text (context, p_result, len, free);
	    }
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
Srid (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ Srid(BLOB encoded geometry)
/
/ returns the SRID
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
	sqlite3_result_int (context, sple_get_srid (p_blob));
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
IsEmpty (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ IsEmpty(BLOB encoded geometry)
/
/ returns:
/ 1 if this geometry contains no entities
/ 0 otherwise
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int geoms;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_int (context, 1);
    else
      {
	  geoms = sple_count_geometries (geo);
	  if (geoms == 0)
	      sqlite3_result_int (context, 1);
	  else
	      sqlite3_result_int (context, 0);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
Envelope (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ Envelope(BLOB encoded geometry)
/
/ returns the MBR for current geometry 
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int len;
    unsigned char *p_result = NULL;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  sple_envelope (geo, &p_result, &len);
	  if (!p_result)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_blob (context, p_result, len, free);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
X (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ X(BLOB encoded POINT)
/
/ returns the X coordinate for current POINT geometry 
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int ret;
    double coord;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  ret = sple_x (geo, &coord);
	  if (!ret)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_double (context, coord);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
Y (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ Y(BLOB encoded POINT)
/
/ returns the Y coordinate for current POINT geometry 
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int ret;
    double coord;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  ret = sple_y (geo, &coord);
	  if (!ret)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_double (context, coord);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
NumPoints (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ NumPoints(BLOB encoded LINESTRING)
/
/ returns the numer of vertices for current LINESTRING geometry 
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int ret;
    int points;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  ret = sple_num_points (geo, &points);
	  if (!ret)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_int (context, points);
      }
    sple_free_geometry (geo);
}

/*
/ the following functions simply readdress the request to point_n()
/ setting the appropriate request mode
*/

SPATIALITE_DECLARE void
StartPoint (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    point_n (context, argc, argv, SPLE_START_POINT);
}

SPATIALITE_DECLARE void
EndPoint (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    point_n (context, argc, argv, SPLE_END_POINT);
}

SPATIALITE_DECLARE void
PointN (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    point_n (context, argc, argv, SPLE_POINTN);
}

SPATIALITE_DECLARE void
point_n (sqlite3_context * context, int argc, sqlite3_value ** argv,
	 int request)
{
/* SQL functions:
/ StartPoint(BLOB encoded LINESTRING geometry)
/ EndPoint(BLOB encoded LINESTRING geometry)
/ PointN(BLOB encoded LINESTRING geometry, integer point_no)
/
/ returns the Nth POINT for current LINESTRING geometry 
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int vertex;
    int len;
    unsigned char *p_result = NULL;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    if (request == SPLE_POINTN)
      {
	  /* PointN() requires point index to be defined as an SQL function argument */
	  if (sqlite3_value_type (argv[1]) != SQLITE_INTEGER)
	    {
		sqlite3_result_null (context);
		return;
	    }
	  vertex = sqlite3_value_int (argv[1]);
      }
    else if (request == SPLE_END_POINT)
	vertex = -1;		/* EndPoint() specifies a negative point index */
    else
	vertex = 1;		/* StartPoint() */
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  sple_point_n (geo, vertex, &p_result, &len);
	  if (!p_result)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_blob (context, p_result, len, free);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
IsClosed (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ IsClosed(BLOB encoded LINESTRING geometry)
/
/ returns:
/ 1 if this LINESTRING or MULTILINESTRING is closed
/ 0 otherwise
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int ret;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  ret = sple_is_closed (geo);
	  if (ret < 0)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_int (context, ret);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
IsRing (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ IsRing(BLOB encoded LINESTRING geometry)
/
/ returns:
/ 1 if this LINESTRING or MULTILINESTRING is simple AND closed
/ 0 otherwise
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int closed = 0;
    int simple = 0;
    int ret;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  closed = sple_is_closed (geo);
	  if (closed == 1)
	      simple = sple_is_simple (geo);
	  if (closed < 0)
	      sqlite3_result_null (context);
	  else
	    {
		if (simple == 1 && closed == 1)
		    ret = 1;
		else
		    ret = 0;
		sqlite3_result_int (context, ret);
	    }
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
IsSimple (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ IsSimple(BLOB encoded GEOMETRY)
/
/ returns:
/ 1 if this GEOMETRY is simple
/ 0 otherwise
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int ret;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  ret = sple_is_simple (geo);
	  if (ret < 0)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_int (context, ret);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
NumInteriorRings (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ NumInteriorRings(BLOB encoded POLYGON)
/
/ returns the numer of internal rigs for current POLYGON geometry 
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int ret;
    int borders;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  ret = sple_num_interior_rings (geo, &borders);
	  if (!ret)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_int (context, borders);
      }
    sple_free_geometry (geo);
}

/*
/ the following functions simply readdress the request to ring_n()
/ setting the appropriate request mode
*/

SPATIALITE_DECLARE void
ExteriorRing (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    ring_n (context, argc, argv, SPLE_EXTERIOR_RING);
}

SPATIALITE_DECLARE void
InteriorRingN (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    ring_n (context, argc, argv, SPLE_INTERIOR_RING);
}

SPATIALITE_DECLARE void
ring_n (sqlite3_context * context, int argc, sqlite3_value ** argv, int request)
{
/* SQL functions:
/ ExteriorRingN(BLOB encoded POLYGON geometry)
/ InteriorRingN(BLOB encoded POLYGON geometry, integer ring_no)
/
/ returns the Nth RING for current POLYGON geometry 
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int border;
    int len;
    unsigned char *p_result = NULL;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    if (request == SPLE_INTERIOR_RING)
      {
	  if (sqlite3_value_type (argv[1]) != SQLITE_INTEGER)
	    {
		sqlite3_result_null (context);
		return;
	    }
	  border = sqlite3_value_int (argv[1]);
      }
    else
	border = -1;		/* ExteriorRing() defines a negative ring index */
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  sple_ring_n (geo, border, &p_result, &len);
	  if (!p_result)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_blob (context, p_result, len, free);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
NumGeometries (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ NumGeometries(BLOB encoded MULTIxxx or GEOMETRYCOLLECTION)
/
/ returns the numer of geometries for current geometry 
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int ret;
    int geoms;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  ret = sple_num_geometries (geo, &geoms);
	  if (!ret)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_int (context, geoms);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
GeometryN (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ GeometryN(BLOB encoded GEOMETRYCOLLECTION geometry)
/
/ returns the Nth geometry for current GEOMETRYCOLLECTION or MULTIxxxx geometry 
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int entity;
    int len;
    unsigned char *p_result = NULL;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    if (sqlite3_value_type (argv[1]) != SQLITE_INTEGER)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    entity = sqlite3_value_int (argv[1]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  sple_geometry_n (geo, entity, &p_result, &len);
	  if (!p_result)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_blob (context, p_result, len, free);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
Length (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ Length(BLOB encoded LINESTRING or MULTILINESTRING)
/
/ returns the length for current LINESTRING geometry,
/ or the total length for current MULTILINSTRING geometry 
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int ret;
    double length;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  ret = sple_length (geo, &length);
	  if (!ret)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_double (context, length);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
Area (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ Area(BLOB encoded POLYGON or MULTIPOLYGON)
/
/ returns the area for current POLYGON geometry,
/ or the total area for current MULTIPOLYGON geometry 
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int ret;
    double area;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  ret = sple_area (geo, &area);
	  if (!ret)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_double (context, area);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
Centroid (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ Centroid(BLOBencoded POLYGON or MULTIPOLYGON geometry)
/
/ returns a POINT representing the centroid for current POLYGON / MULTIPOLYGON geometry 
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int len;
    unsigned char *p_result = NULL;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  sple_centroid (geo, &p_result, &len);
	  if (!p_result)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_blob (context, p_result, len, free);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
ShiftCoords (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ ShiftCoords(BLOBencoded geometry, shiftX, shiftY)
/
/ returns a new geometry that is the original one received, but with shifted coordinates
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int len;
    unsigned char *p_result = NULL;
    sple_geometry_ptr geo = NULL;
    double shift_x;
    double shift_y;
    int int_value;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    if (sqlite3_value_type (argv[1]) == SQLITE_FLOAT)
	shift_x = sqlite3_value_double (argv[1]);
    else if (sqlite3_value_type (argv[1]) == SQLITE_INTEGER)
      {
	  int_value = sqlite3_value_int (argv[1]);
	  shift_x = int_value;
      }
    else
      {
	  sqlite3_result_null (context);
	  return;
      }
    if (sqlite3_value_type (argv[2]) == SQLITE_FLOAT)
	shift_y = sqlite3_value_double (argv[2]);
    else if (sqlite3_value_type (argv[2]) == SQLITE_INTEGER)
      {
	  int_value = sqlite3_value_int (argv[2]);
	  shift_y = int_value;
      }
    else
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  sple_shift_coords (geo, shift_x, shift_y, &p_result, &len);
	  if (!p_result)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_blob (context, p_result, len, free);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
ScaleCoords (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ ScaleCoords(BLOBencoded geometry, scale_factor_x [, scale_factor_y])
/
/ returns a new geometry that is the original one received, but with scaled coordinates
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int len;
    unsigned char *p_result = NULL;
    sple_geometry_ptr geo = NULL;
    double scale_x;
    double scale_y;
    int int_value;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    if (sqlite3_value_type (argv[1]) == SQLITE_FLOAT)
	scale_x = sqlite3_value_double (argv[1]);
    else if (sqlite3_value_type (argv[1]) == SQLITE_INTEGER)
      {
	  int_value = sqlite3_value_int (argv[1]);
	  scale_x = int_value;
      }
    else
      {
	  sqlite3_result_null (context);
	  return;
      }
    if (argc == 2)
	scale_y = scale_x;	/* this one is an isotropic scaling request */
    else
      {
	  /* an anisotropic scaling is requested */
	  if (sqlite3_value_type (argv[2]) == SQLITE_FLOAT)
	      scale_y = sqlite3_value_double (argv[2]);
	  else if (sqlite3_value_type (argv[2]) == SQLITE_INTEGER)
	    {
		int_value = sqlite3_value_int (argv[2]);
		scale_y = int_value;
	    }
	  else
	    {
		sqlite3_result_null (context);
		return;
	    }
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  sple_scale_coords (geo, scale_x, scale_y, &p_result, &len);
	  if (!p_result)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_blob (context, p_result, len, free);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
RotateCoords (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ RotateCoords(BLOBencoded geometry, angle)
/
/ returns a new geometry that is the original one received, but with rotated coordinates
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int len;
    unsigned char *p_result = NULL;
    sple_geometry_ptr geo = NULL;
    double angle;
    int int_value;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    if (sqlite3_value_type (argv[1]) == SQLITE_FLOAT)
	angle = sqlite3_value_double (argv[1]);
    else if (sqlite3_value_type (argv[1]) == SQLITE_INTEGER)
      {
	  int_value = sqlite3_value_int (argv[1]);
	  angle = int_value;
      }
    else
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  sple_rotate_coords (geo, angle, &p_result, &len);
	  if (!p_result)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_blob (context, p_result, len, free);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
ReflectCoords (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ ReflectCoords(BLOBencoded geometry, x_axis,  y_axis)
/
/ returns a new geometry that is the original one received, but with mirrored coordinates
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int len;
    unsigned char *p_result = NULL;
    sple_geometry_ptr geo = NULL;
    int x_axis;
    int y_axis;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    if (sqlite3_value_type (argv[1]) == SQLITE_INTEGER)
	x_axis = sqlite3_value_int (argv[1]);
    else
      {
	  sqlite3_result_null (context);
	  return;
      }
    if (sqlite3_value_type (argv[2]) == SQLITE_INTEGER)
	y_axis = sqlite3_value_int (argv[2]);
    else
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  sple_reflect_coords (geo, x_axis, y_axis, &p_result, &len);
	  if (!p_result)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_blob (context, p_result, len, free);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
SetSrid (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ SetSrid(BLOBencoded geometry, srid)
/
/ returns a new geometry that is the original one received, but with the new SRID [no coordinates translation is applied]
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    sple_geometry_ptr geo = NULL;
    int srid;
    unsigned char *p_result = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    if (sqlite3_value_type (argv[1]) == SQLITE_INTEGER)
	srid = sqlite3_value_int (argv[1]);
    else
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  p_result = malloc (n_bytes);
	  memcpy (p_result, p_blob, n_bytes);
	  sple_set_srid (p_result, srid);
	  sqlite3_result_blob (context, p_result, n_bytes, free);
      }
    sple_free_geometry (geo);
}

SPATIALITE_DECLARE void
Transform (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
/* SQL function:
/ Transform(BLOBencoded geometry, srid)
/
/ returns a new geometry that is the original one received, but with the new SRID [no coordinates translation is applied]
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    sple_geometry_ptr geo = NULL;
    int srid_from;
    int srid_to;
    char proj_from[256];
    char proj_to[256];
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    if (sqlite3_value_type (argv[1]) == SQLITE_INTEGER)
	srid_to = sqlite3_value_int (argv[1]);
    else
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_blob_decode (p_blob, n_bytes);
    if (!geo)
	sqlite3_result_null (context);
    else
      {
	  *proj_from = '\0';
	  *proj_to = '\0';
	  srid_from = sple_get_srid (p_blob);
	  sple_get_proj_params (srid_from, proj_from);
	  sple_get_proj_params (srid_to, proj_to);
	  if (*proj_to == '\0' || *proj_from == '\0')
	    {
		sple_free_geometry (geo);
		sqlite3_result_null (context);
		return;
	    }
	  /* constructs the PROJ4 params */
	  int from_angle = sple_is_longlat (proj_from);
	  int to_angle = sple_is_longlat (proj_to);
	  projPJ from_cs = pj_init_plus (proj_from);
	  projPJ to_cs = pj_init_plus (proj_to);
	  int ie;
	  int ip;
	  long pts = 0;
	  sple_generic_geometry_ptr pGeo;
	  sple_point_ptr point = NULL;
	  sple_linestring_ptr line = NULL;
	  sple_polygon_ptr polyg = NULL;
	  double *x = NULL;
	  double *y = NULL;
	  double *z = NULL;
	  int error = 0;
	  for (ie = 0; ie < geo->entities; ie++)
	    {
		/* how many entities, and of what kind, do we have ? */
		pGeo = *(geo->geometries + ie);
		switch (pGeo->type)
		  {
		      /* loops on individual geometric entities */
		  case SPLE_POINT:
		      pts = 1;
		      break;
		  case SPLE_LINESTRING:
		      line = (sple_linestring_ptr) (pGeo->geometry);
		      pts = line->points;
		      break;
		  case SPLE_POLYGON:
		      polyg = (sple_polygon_ptr) (pGeo->geometry);
		      pts = polyg->external->points;
		      break;
		  };
		if (pts > 0)
		  {
		      /* allocating the coordinate arrays */
		      x = malloc (sizeof (double) * pts);
		      y = malloc (sizeof (double) * pts);
		      z = malloc (sizeof (double) * pts);
		      switch (pGeo->type)
			{
			    /* prepares the coordinate arrays */
			case SPLE_POINT:
			    point = (sple_point_ptr) (pGeo->geometry);
			    if (from_angle)
			      {
				  x[0] = sple_degs_to_rads (point->x);
				  y[0] = sple_degs_to_rads (point->y);
			      }
			    else
			      {
				  x[0] = point->x;
				  y[0] = point->y;
			      }
			    z[0] = 0.0;
			    break;
			case SPLE_LINESTRING:
			    line = (sple_linestring_ptr) (pGeo->geometry);
			    for (ip = 0; ip < line->points; ip++)
			      {
				  if (from_angle)
				    {
					x[ip] =
					    sple_degs_to_rads ((line->geometry +
								ip)->x);
					y[ip] =
					    sple_degs_to_rads ((line->geometry +
								ip)->y);
				    }
				  else
				    {
					x[ip] = (line->geometry + ip)->x;
					y[ip] = (line->geometry + ip)->y;
				    }
				  z[ip] = 0.0;
			      }
			    break;
			case SPLE_POLYGON:
			    polyg = (sple_polygon_ptr) (pGeo->geometry);
			    for (ip = 0; ip < polyg->external->points; ip++)
			      {
				  if (from_angle)
				    {
					x[ip] =
					    sple_degs_to_rads ((polyg->
								external->
								geometry +
								ip)->x);
					y[ip] =
					    sple_degs_to_rads ((polyg->
								external->
								geometry +
								ip)->y);
				    }
				  else
				    {
					x[ip] =
					    (polyg->external->geometry + ip)->x;
					y[ip] =
					    (polyg->external->geometry + ip)->y;
				    }
				  z[ip] = 0.0;
			      }
			    break;
			};
		      if (pj_transform (from_cs, to_cs, pts, 0, x, y, z) == 0)
			{
			    /* sets new coordinate values */
			    switch (pGeo->type)
			      {
				  /* prepares the coordinate arrays */
			      case SPLE_POINT:
				  point = (sple_point_ptr) (pGeo->geometry);
				  if (to_angle)
				    {
					point->x = sple_rads_to_degs (x[0]);
					point->y = sple_rads_to_degs (y[0]);
				    }
				  else
				    {
					point->x = x[0];
					point->y = y[0];
				    }
				  break;
			      case SPLE_LINESTRING:
				  line = (sple_linestring_ptr) (pGeo->geometry);
				  for (ip = 0; ip < line->points; ip++)
				    {
					if (to_angle)
					  {
					      (line->geometry + ip)->x =
						  sple_rads_to_degs (x[ip]);
					      (line->geometry + ip)->y =
						  sple_rads_to_degs (y[ip]);
					  }
					else
					  {
					      (line->geometry + ip)->x = x[ip];
					      (line->geometry + ip)->y = y[ip];
					  }
				    }
				  break;
			      case SPLE_POLYGON:
				  polyg = (sple_polygon_ptr) (pGeo->geometry);
				  for (ip = 0; ip < polyg->external->points;
				       ip++)
				    {
					if (to_angle)
					  {
					      (polyg->external->geometry +
					       ip)->x =
		      sple_rads_to_degs (x[ip]);
					      (polyg->external->geometry +
					       ip)->y =
		      sple_rads_to_degs (y[ip]);
					  }
					else
					  {
					      (polyg->external->geometry +
					       ip)->x = x[ip];
					      (polyg->external->geometry +
					       ip)->y = y[ip];
					  }
				    }
				  break;
			      };
			}
		      else
			  error = 1;
		      /* destroying the coordinate arrays */
		      free (x);
		      free (y);
		      free (z);
		      if (pGeo->type == SPLE_POLYGON)
			{
			    /* for POLYGONs we need to process the INTERIOR RINGs as well */
			    int ib;
			    polyg = (sple_polygon_ptr) (pGeo->geometry);
			    for (ib = 0; ib < polyg->borders; ib++)
			      {
				  line = *(polyg->internals + ib);
				  pts = line->points;
				  if (pts > 0)
				    {
					/* allocating the coordinate arrays */
					x = malloc (sizeof (double) * pts);
					y = malloc (sizeof (double) * pts);
					z = malloc (sizeof (double) * pts);
					for (ip = 0; ip < line->points; ip++)
					  {
					      if (from_angle)
						{
						    x[ip] =
							sple_degs_to_rads ((line->geometry + ip)->x);
						    y[ip] =
							sple_degs_to_rads ((line->geometry + ip)->y);
						}
					      else
						{
						    x[ip] =
							(line->geometry +
							 ip)->x;
						    y[ip] =
							(line->geometry +
							 ip)->y;
						}
					      z[ip] = 0.0;
					  }
					if (pj_transform
					    (from_cs, to_cs, pts, 0, x, y,
					     z) == 0)
					  {
					      /* sets new coordinate values */
					      for (ip = 0; ip < line->points;
						   ip++)
						{
						    if (to_angle)
						      {
							  (line->geometry +
							   ip)->x =
					     sple_rads_to_degs (x[ip]);
							  (line->geometry +
							   ip)->y =
					     sple_rads_to_degs (y[ip]);
						      }
						    else
						      {
							  (line->geometry +
							   ip)->x = x[ip];
							  (line->geometry +
							   ip)->y = y[ip];
						      }
						}
					  }
					/* destroying the coordinate arrays */
					free (x);
					free (y);
					free (z);
				    }
			      }
			}
		  }
	    }
	  /* destroys the PROJ4 params */
	  pj_free (from_cs);
	  pj_free (to_cs);
	  if (error)
	      sqlite3_result_null (context);
	  else
	    {
		/* builds the BLOB geometry to be returned */
		int len;
		unsigned char *p_result = NULL;
		sple_blob_encode (geo, &p_result, &len);
		sple_set_srid (p_result, srid_to);
		sqlite3_result_blob (context, p_result, len, free);
	    }
      }
    sple_free_geometry (geo);
}

/*
/ the following functions simply readdress the request to geom_from_text?()
/ setting the appropriate GEOMETRY CLASS TYPE
*/

SPATIALITE_DECLARE void
GeomFromText1 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_text1 (context, argc, argv, (short) -1);
}

SPATIALITE_DECLARE void
GeomFromText2 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_text2 (context, argc, argv, (short) -1);
}

SPATIALITE_DECLARE void
GeomCollFromText1 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_text1 (context, argc, argv, (short) SPLE_GEOMETRYCOLLECTION);
}

SPATIALITE_DECLARE void
GeomCollFromText2 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_text2 (context, argc, argv, (short) SPLE_GEOMETRYCOLLECTION);
}

SPATIALITE_DECLARE void
LineFromText1 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_text1 (context, argc, argv, (short) SPLE_LINESTRING);
}

SPATIALITE_DECLARE void
LineFromText2 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_text2 (context, argc, argv, (short) SPLE_LINESTRING);
}

SPATIALITE_DECLARE void
PointFromText1 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_text1 (context, argc, argv, (short) SPLE_POINT);
}

SPATIALITE_DECLARE void
PointFromText2 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_text2 (context, argc, argv, (short) SPLE_POINT);
}

SPATIALITE_DECLARE void
PolyFromText1 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_text1 (context, argc, argv, (short) SPLE_POLYGON);
}

SPATIALITE_DECLARE void
PolyFromText2 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_text2 (context, argc, argv, (short) SPLE_POLYGON);
}

SPATIALITE_DECLARE void
MLineFromText1 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_text1 (context, argc, argv, (short) SPLE_MULTILINESTRING);
}

SPATIALITE_DECLARE void
MLineFromText2 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_text2 (context, argc, argv, (short) SPLE_MULTILINESTRING);
}

SPATIALITE_DECLARE void
MPointFromText1 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_text1 (context, argc, argv, (short) SPLE_MULTIPOINT);
}

SPATIALITE_DECLARE void
MPointFromText2 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_text2 (context, argc, argv, (short) SPLE_MULTIPOINT);
}

SPATIALITE_DECLARE void
MPolyFromText1 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_text1 (context, argc, argv, (short) SPLE_MULTIPOLYGON);
}

SPATIALITE_DECLARE void
MPolyFromText2 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_text2 (context, argc, argv, (short) SPLE_MULTIPOLYGON);
}

SPATIALITE_DECLARE void
geom_from_text1 (sqlite3_context * context, int argc, sqlite3_value ** argv,
		 short type)
{
/* SQL function:
/ GeomFromText(WKT encoded geometry)
/
/ returns the current geometry by parsing WKT encoded string 
/ or NULL if any error is encountered
/
/ if *type* is a negative value can accept any GEOMETRY CLASS
/ otherwise only requests conforming with required CLASS are valid
*/
    int len;
    unsigned char *p_result = NULL;
    const unsigned char *text;
    char *clean;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_TEXT)
      {
	  sqlite3_result_null (context);
	  return;
      }
    text = sqlite3_value_text (argv[0]);
    clean = sple_clean_wkt (text);
    if (!clean)
      {
	  sqlite3_result_null (context);
	  return;
      }
    geo = sple_parse_wkt (clean, type);
    if (geo == NULL)
      {
	  free (clean);
	  sqlite3_result_null (context);
	  return;
      }
    free (clean);
    sple_blob_encode (geo, &p_result, &len);
    sple_free_geometry (geo);
    sqlite3_result_blob (context, p_result, len, free);
}

SPATIALITE_DECLARE void
geom_from_text2 (sqlite3_context * context, int argc, sqlite3_value ** argv,
		 short type)
{
/* SQL function:
/ GeomFromText(WKT encoded geometry, SRID)
/
/ returns the current geometry by parsing WKT encoded string 
/ or NULL if any error is encountered
/
/ if *type* is a negative value can accept any GEOMETRY CLASS
/ otherwise only requests conforming with required CLASS are valid
*/
    int len;
    unsigned char *p_result = NULL;
    const unsigned char *text;
    char *clean;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_TEXT)
      {
	  sqlite3_result_null (context);
	  return;
      }
    if (sqlite3_value_type (argv[1]) != SQLITE_INTEGER)
      {
	  sqlite3_result_null (context);
	  return;
      }
    text = sqlite3_value_text (argv[0]);
    clean = sple_clean_wkt (text);
    if (!clean)
      {
	  sqlite3_result_null (context);
	  return;
      }
    geo = sple_parse_wkt (clean, type);
    if (geo == NULL)
      {
	  free (clean);
	  sqlite3_result_null (context);
	  return;
      }
    free (clean);
    sple_blob_encode (geo, &p_result, &len);
    sple_set_srid (p_result, sqlite3_value_int (argv[1]));
    sple_free_geometry (geo);
    sqlite3_result_blob (context, p_result, len, free);
}

/*
/ the following functions simply readdress the request to geom_from_wkb?()
/ setting the appropriate GEOMETRY CLASS TYPE
*/

SPATIALITE_DECLARE void
GeomFromWKB1 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_wkb1 (context, argc, argv, (short) -1);
}

SPATIALITE_DECLARE void
GeomFromWKB2 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_wkb2 (context, argc, argv, (short) -1);
}

SPATIALITE_DECLARE void
GeomCollFromWKB1 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_wkb1 (context, argc, argv, (short) SPLE_GEOMETRYCOLLECTION);
}

SPATIALITE_DECLARE void
GeomCollFromWKB2 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_wkb2 (context, argc, argv, (short) SPLE_GEOMETRYCOLLECTION);
}

SPATIALITE_DECLARE void
LineFromWKB1 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_wkb1 (context, argc, argv, (short) SPLE_LINESTRING);
}

SPATIALITE_DECLARE void
LineFromWKB2 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_wkb2 (context, argc, argv, (short) SPLE_LINESTRING);
}

SPATIALITE_DECLARE void
PointFromWKB1 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_wkb1 (context, argc, argv, (short) SPLE_POINT);
}

SPATIALITE_DECLARE void
PointFromWKB2 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_wkb2 (context, argc, argv, (short) SPLE_POINT);
}

SPATIALITE_DECLARE void
PolyFromWKB1 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_wkb1 (context, argc, argv, (short) SPLE_POLYGON);
}

SPATIALITE_DECLARE void
PolyFromWKB2 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_wkb2 (context, argc, argv, (short) SPLE_POLYGON);
}

SPATIALITE_DECLARE void
MLineFromWKB1 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_wkb1 (context, argc, argv, (short) SPLE_MULTILINESTRING);
}

SPATIALITE_DECLARE void
MLineFromWKB2 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_wkb2 (context, argc, argv, (short) SPLE_MULTILINESTRING);
}

SPATIALITE_DECLARE void
MPointFromWKB1 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_wkb1 (context, argc, argv, (short) SPLE_MULTIPOINT);
}

SPATIALITE_DECLARE void
MPointFromWKB2 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_wkb2 (context, argc, argv, (short) SPLE_MULTIPOINT);
}

SPATIALITE_DECLARE void
MPolyFromWKB1 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_wkb1 (context, argc, argv, (short) SPLE_MULTIPOLYGON);
}

SPATIALITE_DECLARE void
MPolyFromWKB2 (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    geom_from_wkb2 (context, argc, argv, (short) SPLE_MULTIPOLYGON);
}

SPATIALITE_DECLARE void
geom_from_wkb1 (sqlite3_context * context, int argc, sqlite3_value ** argv,
		short type)
{
/* SQL function:
/ GeomFromWKB(WKB encoded geometry)
/
/ returns the current geometry by parsing WKB encoded blob 
/ or NULL if any error is encountered
/
/ if *type* is a negative value can accept any GEOMETRY CLASS
/ otherwise only requests conforming with required CLASS are valid
*/
    int len;
    int n_bytes;
    unsigned char *p_result = NULL;
    const unsigned char *wkb;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    wkb = sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_parse_wkb (wkb, n_bytes, type);
    if (geo == NULL)
      {
	  sqlite3_result_null (context);
	  return;
      }
    sple_blob_encode (geo, &p_result, &len);
    sple_free_geometry (geo);
    sqlite3_result_blob (context, p_result, len, free);
}

SPATIALITE_DECLARE void
geom_from_wkb2 (sqlite3_context * context, int argc, sqlite3_value ** argv,
		short type)
{
/* SQL function:
/ GeomFromWKB(WKB encoded geometry, SRID)
/
/ returns the current geometry by parsing WKB encoded string 
/ or NULL if any error is encountered
/
/ if *type* is a negative value can accept any GEOMETRY CLASS
/ otherwise only requests conforming with required CLASS are valid
*/
    int len;
    int n_bytes;
    unsigned char *p_result = NULL;
    const unsigned char *wkb;
    sple_geometry_ptr geo = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    if (sqlite3_value_type (argv[1]) != SQLITE_INTEGER)
      {
	  sqlite3_result_null (context);
	  return;
      }
    wkb = sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo = sple_parse_wkb (wkb, n_bytes, type);
    if (geo == NULL)
      {
	  sqlite3_result_null (context);
	  return;
      }
    sple_blob_encode (geo, &p_result, &len);
    sple_set_srid (p_result, sqlite3_value_int (argv[1]));
    sple_free_geometry (geo);
    sqlite3_result_blob (context, p_result, len, free);
}

/*
/ the following functions simply readdress the mbr_eval()
/ setting the appropriate request mode
*/

SPATIALITE_DECLARE void
MBRContains (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    mbrs_eval (context, argc, argv, SPLE_MBR_CONTAINS);
}

SPATIALITE_DECLARE void
MBRDisjoint (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    mbrs_eval (context, argc, argv, SPLE_MBR_DISJOINT);
}

SPATIALITE_DECLARE void
MBREqual (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    mbrs_eval (context, argc, argv, SPLE_MBR_EQUAL);
}

SPATIALITE_DECLARE void
MBRIntersects (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    mbrs_eval (context, argc, argv, SPLE_MBR_INTERSECTS);
}

SPATIALITE_DECLARE void
MBROverlaps (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    mbrs_eval (context, argc, argv, SPLE_MBR_OVERLAPS);
}

SPATIALITE_DECLARE void
MBRTouches (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    mbrs_eval (context, argc, argv, SPLE_MBR_TOUCHES);
}

SPATIALITE_DECLARE void
MBRWithin (sqlite3_context * context, int argc, sqlite3_value ** argv)
{
    mbrs_eval (context, argc, argv, SPLE_MBR_WITHIN);
}

SPATIALITE_DECLARE void
mbrs_eval (sqlite3_context * context, int argc, sqlite3_value ** argv,
	   int request)
{
/* SQL function:
/ MBRsomething(BLOB encoded GEOMETRY-1, BLOB encoded GEOMETRY-2)
/
/ returns:
/ 1 if the required spatial relationship between the two MBRs is TRUE
/ 0 otherwise
/ or NULL if any error is encountered
*/
    unsigned char *p_blob;
    int n_bytes;
    int ret;
    sple_geometry_ptr geo1 = NULL;
    sple_geometry_ptr geo2 = NULL;
    if (sqlite3_value_type (argv[0]) != SQLITE_BLOB)
      {
	  sqlite3_result_null (context);
	  return;
      }
    p_blob = (unsigned char *) sqlite3_value_blob (argv[0]);
    n_bytes = sqlite3_value_bytes (argv[0]);
    geo1 = sple_mbr_decode (p_blob, n_bytes);
    p_blob = (unsigned char *) sqlite3_value_blob (argv[1]);
    n_bytes = sqlite3_value_bytes (argv[1]);
    geo2 = sple_mbr_decode (p_blob, n_bytes);
    if (!geo1 || !geo2)
	sqlite3_result_null (context);
    else
      {
	  ret = 0;
	  switch (request)
	    {
	    case SPLE_MBR_CONTAINS:
		ret =
		    sple_contains_mbrs (geo1->minx, geo1->miny, geo1->maxx,
					geo1->maxy, geo2->minx, geo2->miny,
					geo2->maxx, geo2->maxy);
		break;
	    case SPLE_MBR_DISJOINT:
		ret =
		    sple_disjoint_mbrs (geo1->minx, geo1->miny, geo1->maxx,
					geo1->maxy, geo2->minx, geo2->miny,
					geo2->maxx, geo2->maxy);
		break;
	    case SPLE_MBR_EQUAL:
		ret =
		    sple_equal_mbrs (geo1->minx, geo1->miny, geo1->maxx,
				     geo1->maxy, geo2->minx, geo2->miny,
				     geo2->maxx, geo2->maxy);
		break;
	    case SPLE_MBR_INTERSECTS:
		ret =
		    sple_intersects_mbrs (geo1->minx, geo1->miny, geo1->maxx,
					  geo1->maxy, geo2->minx, geo2->miny,
					  geo2->maxx, geo2->maxy);
		break;
	    case SPLE_MBR_OVERLAPS:
		ret =
		    sple_overlaps_mbrs (geo1->minx, geo1->miny, geo1->maxx,
					geo1->maxy, geo2->minx, geo2->miny,
					geo2->maxx, geo2->maxy);
		break;
	    case SPLE_MBR_TOUCHES:
		ret =
		    sple_touches_mbrs (geo1->minx, geo1->miny, geo1->maxx,
				       geo1->maxy, geo2->minx, geo2->miny,
				       geo2->maxx, geo2->maxy);
		break;
	    case SPLE_MBR_WITHIN:
		ret =
		    sple_within_mbrs (geo1->minx, geo1->miny, geo1->maxx,
				      geo1->maxy, geo2->minx, geo2->miny,
				      geo2->maxx, geo2->maxy);
		break;
	    }
	  if (ret == 0)
	      sqlite3_result_null (context);
	  else
	      sqlite3_result_int (context, ret);
      }
    sple_free_geometry (geo1);
    sple_free_geometry (geo2);
}

static void
sple_get_proj_params (int srid, char *proj_params)
{
/* retrives the PROJ params from EPSG table, if possible */
    char sql[256];
    char **results;
    int rows;
    int columns;
    int i;
    char *errMsg = NULL;
    *proj_params = '\0';
    sprintf (sql, "SELECT proj_params FROM epsg WHERE srid = %d", srid);
    int ret =
	sqlite3_get_table (this_db, sql, &results, &rows, &columns, &errMsg);
    if (ret != SQLITE_OK)
      {
	  fprintf (stderr, "unknown SRID: %d\t<%s>\n", srid, errMsg);
	  sqlite3_free (errMsg);
	  return;
      }
    for (i = 1; i <= rows; i++)
	strcpy (proj_params, results[(i * columns)]);
    if (*proj_params == '\0')
	fprintf (stderr, "unknown SRID: %d\n", srid);
}

static int
sple_is_longlat (char *str)
{
/* checks if we have to do with ANGLES if +proj=longlat is defined */
    if (strstr (str, "+proj=longlat") != NULL)
	return 1;
    return 0;
}

static double
sple_rads_to_degs (double rads)
{
/* converts an ANGLE from radians to degrees */
    return rads * RAD_TO_DEG;
}

static double
sple_degs_to_rads (double degs)
{
/* converts an ANGLE from degrees to radians */
    return degs * DEG_TO_RAD;
}
